# Generated by Django 3.2.18 on 2023-05-15 12:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('location', '0013_auto_20230317_1534'),
        ('tools', '0005_items_services_rights'),
    ]

    operations = [
        migrations.AddField(
            model_name='extract',
            name='health_facility',
            field=models.ForeignKey(db_column='HFID', on_delete=django.db.models.deletion.DO_NOTHING, to='location.healthfacility'),
        ),
        migrations.AddField(
            model_name='extract',
            name='location',
            field=models.ForeignKey(db_column='LocationId', on_delete=django.db.models.deletion.DO_NOTHING, to='location.location'),
        ),
    ]
