# Generated by Django 3.2.16 on 2022-12-07 17:25
from core.utils import insert_role_right_for_system
from django.db import migrations


def add_rights(apps, schema_editor):
    insert_role_right_for_system(64, 131007)  # items
    insert_role_right_for_system(64, 131008)  # items
    insert_role_right_for_system(64, 131009)  # services
    insert_role_right_for_system(64, 131010)  # services


class Migration(migrations.Migration):

    dependencies = [
        ('tools', '0004_registers_right_for_scheme_admin'),
    ]

    operations = [
        migrations.RunPython(add_rights),
    ]
