# Generated by Django 3.0.14 on 2021-12-08 11:17
from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("tools", "0001_initial"),
    ]

    operations = [
        migrations.RunSQL(
            "ALTER TABLE tblExtracts ADD ExtractFile nvarchar(255) NULL"
            if settings.MSSQL else
            'ALTER TABLE "tblExtracts" ADD "ExtractFile" varchar(255)',
            reverse_sql='ALTER TABLE "tblExtracts" DROP "ExtractFile"',
        ),
    ]
